@@ -5,7 +5,7 @@ use warnings;
 use Carp;
 use Cwd 'realpath';
 
-our $VERSION   = '1.3070';
+our $VERSION   = '1.3080';
 our $AUTHORITY = 'SUKRIA';
 
 use Dancer::App;
@@ -67,6 +67,7 @@ our @EXPORT    = qw(
   params
   pass
   path
+  patch
   post
   prefix
   push_header
@@ -99,10 +100,26 @@ our @EXPORT    = qw(
 
 # Dancer's syntax
 
-sub after           { Dancer::Hook->new('after', @_) }
+sub after           {
+    Dancer::Deprecation->deprecated(reason => "use hooks!",
+                                    version => '1.3080',
+                                    fatal => 0);
+    Dancer::Hook->new('after', @_);
+}
+sub before          {
+    Dancer::Deprecation->deprecated(reason => "use hooks!",
+                                    version => '1.3080',
+                                    fatal => 0);
+    Dancer::Hook->new('before', @_);
+}
+sub before_template {
+    Dancer::Deprecation->deprecated(reason => "use hooks!",
+                                    version => '1.3080',
+                                    fatal => 0);
+    Dancer::Hook->new('before_template', @_);
+}
+
 sub any             { Dancer::App->current->registry->any_add(@_) }
-sub before          { Dancer::Hook->new('before', @_) }
-sub before_template { Dancer::Hook->new('before_template', @_) }
 sub captures        { Dancer::SharedData->request->params->{captures} }
 sub cookie          { Dancer::Cookies->cookie( @_ ) }
 sub cookies         { Dancer::Cookies->cookies }
@@ -129,19 +146,20 @@ sub hook            { Dancer::Hook->new(@_) }
 sub layout          {
     Dancer::Deprecation->deprecated(reason => "use 'set layout => \"value\"'",
                                     version => '1.3050',
-                                    fatal => 0);
-    set(layout => shift) }
+                                    fatal => 1);
+}
 sub load            { require $_ for @_ }
 sub load_app        { goto &_load_app } # goto doesn't add a call frame. So caller() will work as expected
 sub logger          {
-    Dancer::Deprecation->deprecated(reason => "use 'set logger'",fatal => 0,version=>'1.3050');
-    set(logger => @_)
+    Dancer::Deprecation->deprecated(reason => "use 'set logger => \"value\"'",
+                                    fatal => 1,version=>'1.3050');
 }
 sub mime            { Dancer::MIME->instance() }
 sub options         { Dancer::App->current->registry->universal_add('options', @_) }
 sub params          { Dancer::SharedData->request->params(@_) }
 sub param           { params->{$_[0]} }
 sub pass            { Dancer::SharedData->response->pass(1) }
+sub patch            { Dancer::App->current->registry->universal_add('patch', @_) }
 sub path            { Dancer::FileUtils::path(@_) }
 sub post            { Dancer::App->current->registry->universal_add('post', @_) }
 sub prefix          { Dancer::App->current->set_prefix(@_) }
@@ -224,7 +242,7 @@ sub _load_app {
     my $app = Dancer::App->set_running_app($app_name);
 
     # Application options
-    $app->prefix($options{prefix})     if $options{prefix};
+    $app->set_app_prefix($options{prefix}) if $options{prefix};
     $app->settings($options{settings}) if $options{settings};
 
     # load the application
@@ -258,10 +276,11 @@ sub _init_script_dir {
 
     my $LAYOUT_PRE_DANCER_1_2 = 1;
 
-    # in bin/ or public/ we need to go one level upper to find the appdir
+    # in bin/ or public/ or t/ we need to go one level upper to find the appdir
     $LAYOUT_PRE_DANCER_1_2 = 0
       if ($script_dirs[$#script_dirs - 1] eq 'bin')
-      or ($script_dirs[$#script_dirs - 1] eq 'public');
+      or ($script_dirs[$#script_dirs - 1] eq 'public')
+      or ($script_dirs[$#script_dirs - 1] eq 't');
 
     my $appdir = $ENV{DANCER_APPDIR} || (
           $LAYOUT_PRE_DANCER_1_2
@@ -316,20 +335,85 @@ sub _session {
 
 sub _send_file {
     my ($path, %options) = @_;
+    my $env = Dancer::SharedData->request->env;
 
     my $request = Dancer::Request->new_for_request('GET' => $path);
     Dancer::SharedData->request($request);
 
+    # if you asked for streaming but it's not supported in PSGI
+    if ( $options{'streaming'} && ! $env->{'psgi.streaming'} ) {
+        # TODO: throw a fit (AKA "exception") or a Dancer::Error (or croak)?
+        croak 'Sorry, streaming is not supported on this server.';
+    }
+
     if (exists($options{content_type})) {
         $request->content_type($options{content_type});
     }
 
     my $resp;
-    if ($options{system_path} && -f $path) {
-        $resp = Dancer::Renderer->get_file_response_for_path($path);
+    if (ref($path) eq "SCALAR") {
+        # send_data
+        $resp = Dancer::SharedData->response() || Dancer::Response->new();
+        $resp->header('Content-Type' => exists($options{content_type}) ?
+                                        $options{content_type} : Dancer::MIME->default());
+        $resp->content($$path);
     } else {
-        $resp = Dancer::Renderer->get_file_response();
+        # real send_file
+        if ($options{system_path} && -f $path) {
+            $resp = Dancer::Renderer->get_file_response_for_path($path);
+        } else {
+            $resp = Dancer::Renderer->get_file_response();
+        }
     }
+
+    if (exists($options{filename})) {
+        $resp->push_header('Content-Disposition' => 
+            "attachment; filename=\"$options{filename}\""
+        );
+    }
+
+    if ( $options{'streaming'} ) {
+        # handle streaming
+        $resp->streamed( sub {
+            my ( $status, $headers ) = @_;
+            my %callbacks = defined $options{'callbacks'} ?
+                            %{ $options{'callbacks'} }    :
+                            {};
+
+            return sub {
+                my $respond = shift;
+                exists $callbacks{'override'}
+                    and return $callbacks{'override'}->( $respond, $resp );
+
+                # get respond callback and set headers, get writer in return
+                my $writer = $respond->( [
+                    $status,
+                    $headers,
+                ] );
+
+                # get content from original response
+                my $content = $resp->content;
+
+                exists $callbacks{'around'}
+                    and return $callbacks{'around'}->( $writer, $content );
+
+                if ( ref $content ) {
+                    my $bytes = $options{'bytes'} || '43008'; # 42K (dams)
+                    my $buf;
+                    while ( ( my $read = sysread $content, $buf, $bytes ) != 0 ) {
+                        if ( exists $callbacks{'around_content'} ) {
+                            $callbacks{'around_content'}->( $writer, $buf );
+                        } else {
+                            $writer->write($buf);
+                        }
+                    }
+                } else {
+                    $writer->write($content);
+                }
+            };
+        } );
+    }
+
     return $resp if $resp;
 
     Dancer::Error->new(
@@ -422,7 +506,7 @@ your app.  You can control the exporting through the normal
 L<Exporter> mechanism.  For example:
 
     # Just export the route controllers
-    use Dancer qw(before after get post);
+    use Dancer qw(before after get post put patch);
 
     # Export everything but pass to avoid clashing with Test::More
     use Test::More;
@@ -469,6 +553,11 @@ This is useful when you want to use your Dancer application from a script.
     use Dancer ':script';
     MyApp::schema('DBSchema')->deploy();
 
+By default, the L<warnings> pragma will also be exported, meaning your
+app/script will be running under C<use warnings>.  If you do not want this, set
+the L<import_warnings|Dancer::Config/import_warnings> setting to a false value.
+
+
 =head1 FUNCTIONS
 
 =head2 after
@@ -477,11 +566,13 @@ Add a hook at the B<after> position:
 
     after sub {
         my $response = shift;
-        # do something with request
+        # do something with response, e.g.:
+        $response->header('X-Beer' => 'Yes please');
     };
 
 The anonymous function which is given to C<after> will be executed after
-having executed a route.
+having executed a route, but before the response is returned (so you can modify
+the response here, if you need to)..
 
 You can define multiple after filters, using the C<after> helper as
 many times as you wish; each filter will be executed, in the order you added
@@ -544,7 +635,7 @@ tokens that will be inserted into the template.
 This filter works as the C<before> and C<after> filter.
 
 Now the preferred way for this is to use C<hook>s (namely, the
-C<before_template> one). Check C<hook> documentation bellow.
+C<before_template> one). Check C<hook> documentation below.
 
 =head2 cookies
 
@@ -620,6 +711,8 @@ Alias for the C<start> keyword.
 Logs a message of debug level:
 
     debug "This is a debug message";
+    
+See L<Dancer::Logger> for details on how to configure where log messages go.
 
 =head2 dirname
 
@@ -641,6 +734,8 @@ Logs a message of error level:
 
     error "This is an error message";
 
+See L<Dancer::Logger> for details on how to configure where log messages go.
+
 =head2 false
 
 Constant that returns a false value (0).
@@ -733,6 +828,9 @@ Defines a route for HTTP B<GET> requests to the given path:
         return "Hello world";
     }
 
+Note that a route to match B<HEAD> requests is automatically created as well.
+
+
 =head2 halt
 
 Sets a response object with the content given.
@@ -783,11 +881,16 @@ Do the same as C<header>, but allow for multiple headers with the same name.
 
 Adds a hook at some position. For example :
 
-  hook before_serialization => sub {
+  hook before_serializer => sub {
     my $response = shift;
     $response->content->{generated_at} = localtime();
   };
 
+There can be multiple hooks assigned to a given position, and each will be
+executed in order.
+
+(For details on how to register new hooks from within plugins, see
+L<Dancer::Hook>.)
 Supported B<before> hooks (in order of execution):
 
 =over
@@ -876,7 +979,7 @@ tokens. The second is a ScalarRef representing the content of the template.
     ...
   };
 
-=item before_serialization
+=item before_serializer
 
 This hook receives as argument a L<Dancer::Response> object.
 
@@ -927,9 +1030,13 @@ by the layout
 
 =item after
 
-This is an alias for 'after'.
+This is an alias for C<after>.
 
-This hook receives as argument a L<Dancer::Response> object.
+This hook runs after a request has been processed, but before the response is
+sent.
+
+It receives a L<Dancer::Response> object, which it can modify
+if it needs to make changes to the response which is about to be sent.
 
   hook after => sub {
     my $response = shift;
@@ -989,10 +1096,6 @@ To load multiple apps repeat load_app:
 The old way of loading multiple apps in one go (load_app 'one', 'two';) is
 deprecated.
 
-=head2 mime_type
-
-Deprecated. See L</mime>.
-
 =head2 mime
 
 Shortcut to access the instance object of L<Dancer::MIME>. You should
@@ -1051,6 +1154,25 @@ You should always C<return> after calling C<pass>:
         }
     };
 
+=head2 patch
+
+Defines a route for HTTP B<PATCH> requests to the given URL:
+
+    patch '/resource' => sub { ... };
+
+(C<PATCH> is a relatively new and not-yet-common HTTP verb, which is intended to
+work as a "partial-PUT", transferring just the changes; please see
+L<http://tools.ietf.org/html/rfc5789|RFC5789> for further details.)
+
+Please be aware that, if you run your app in standalone mode, C<PATCH> requests
+will not reach your app unless you have a new version of L<HTTP::Server::Simple>
+which accepts C<PATCH> as a valid verb.  The current version at time of writing,
+C<0.44>, does not.  A pull request has been submitted to add this support, which
+you can find at:
+
+L<https://github.com/bestpractical/http-server-simple/pull/1>
+
+
 =head2 path
 
 Concatenates multiple paths together, without worrying about the underlying
@@ -1065,7 +1187,7 @@ path exists.
 
 Defines a route for HTTP B<POST> requests to the given URL:
 
-    POST '/' => sub {
+    post '/' => sub {
         return "Hello world";
     }
 
@@ -1187,6 +1309,14 @@ instance, to use a custom layout:
 
 Returns a L<Dancer::Request> object representing the current request.
 
+See the L<Dancer::Request> documention for the methods you can call, for
+example:
+
+    request->referer;         # value of the HTTP referer header
+    request->remote_address;  # user's IP address
+    request->user_agent;      # User-Agent header value
+
+
 =head2 send_error
 
 Returns a HTTP error.  By default the HTTP code returned is 500:
@@ -1211,27 +1341,121 @@ the path of the file must be relative to the B<public> directory unless you use
 the C<system_path> option (see below).
 
     get '/download/:file' => sub {
-        send_file(params->{file});
+        return send_file(params->{file});
     }
 
+Send file supports streaming possibility using PSGI streaming. The server should
+support it but normal streaming is supported on most, if not all.
+
+    get '/download/:file' => sub {
+        return send_file( params->{file}, streaming => 1 );
+    }
+
+You can control what happens using callbacks.
+
+First, C<around_content> allows you to get the writer object and the chunk of
+content read, and then decide what to do with each chunk:
+
+    get '/download/:file' => sub {
+        return send_file(
+            params->{file},
+            streaming => 1,
+            callbacks => {
+                around_content => sub {
+                    my ( $writer, $chunk ) = @_;
+                    $writer->write("* $chunk");
+                },
+            },
+        );
+    }
+
+You can use C<around> to all get all the content (whether a filehandle if it's
+a regular file or a full string if it's a scalar ref) and decide what to do with
+it:
+
+    get '/download/:file' => sub {
+        return send_file(
+            params->{file},
+            streaming => 1,
+            callbacks => {
+                around => sub {
+                    my ( $writer, $content 0 = shift;
+
+                    # we know it's a text file, so we'll just stream
+                    # line by line
+                    while ( my $line = <$content> ) {
+                        $writer->write($line);
+                    }
+                },
+            },
+        );
+    }
+
+Or you could use C<override> to control the entire streaming callback request:
+
+    get '/download/:file' => sub {
+        return send_file(
+            params->{file},
+            streaming => 1,
+            callbacks => {
+                override => sub {
+                    my ( $respond, $response ) = @_;
+
+                    my $writer = $respond->( [ $newstatus, $newheaders ] );
+                    $writer->write("some line");
+                },
+            },
+        );
+    }
+
+You can also set the number of bytes that will be read at a time (default being
+42K bytes) using C<bytes>:
+
+    get '/download/:file' => sub {
+        return send_file(
+            params->{file},
+            streaming => 1,
+            bytes     => 524288, # 512K
+        );
+    };
+
+
 The content-type will be set depending on the current MIME types definition
 (see C<mime> if you want to define your own).
 
 If your filename does not have an extension, or you need to force a
 specific mime type, you can pass it to C<send_file> as follows:
 
-    send_file(params->{file}, content_type => 'image/png');
+    return send_file(params->{file}, content_type => 'image/png');
 
 Also, you can use your aliases or file extension names on
 C<content_type>, like this:
 
-    send_file(params->{file}, content_type => 'png');
+    return send_file(params->{file}, content_type => 'png');
 
 For files outside your B<public> folder, you can use the C<system_path>
 switch. Just bear in mind that its use needs caution as it can be
 dangerous.
 
-   send_file('/etc/passwd', system_path => 1);
+   return send_file('/etc/passwd', system_path => 1);
+
+If you have your data in a scalar variable, C<send_file> can be useful
+as well. Pass a reference to that scalar, and C<send_file> will behave
+as if there was a file with that contents:
+
+   return send_file( \$data, content_type => 'image/png' );
+
+Note that Dancer is unable to guess the content type from the data
+contents. Therefore you might need to set the C<content_type>
+properly. For this kind of usage an attribute named C<filename> can be
+useful.  It is used as the Content-Disposition header, to hint the
+brower about the filename it should use.
+
+   return send_file( \$data, content_type => 'image/png'
+                             filename     => 'onion.png' );
+
+Note that you should always use C<return send_file ...> to stop execution of
+your route handler at that point.
 
 =head2 set
 
@@ -1401,6 +1625,12 @@ For example, to disable the layout for a specific request:
         template 'index.tt', {}, { layout => undef };
     };
 
+Or to request a specific layout, of course:
+
+    get '/user' => sub {
+        template 'user.tt', {}, { layout => 'user' };
+    };
+
 Some tokens are automatically added to your template (C<perl_version>,
 C<dancer_version>, C<settings>, C<request>, C<params>, C<vars> and, if
 you have sessions enabled, C<session>).  Check
@@ -1519,7 +1749,11 @@ chain:
 
 =head2 warning
 
-Logs a warning message through the current logger engine.
+Logs a warning message through the current logger engine:
+
+    warning "This is a warning";
+
+See L<Dancer::Logger> for details on how to configure where log messages go.
 
 =head1 AUTHOR
 
